import logging
import random
import socket
import sys
import time
import select

from pieraknet.handlers.offline_ping import OfflinePingHandler
from pieraknet.handlers.open_connection_request_1 import OpenConnectionRequest1Handler
from pieraknet.handlers.open_connection_request_2 import OpenConnectionRequest2Handler
from pieraknet.packets.offline_ping import OfflinePing
from pieraknet.packets.open_connection_request_1 import OpenConnectionRequest1
from pieraknet.packets.open_connection_request_2 import OpenConnectionRequest2
from pieraknet.protocol_info import ProtocolInfo


class ConnectionNotFound(Exception):
    pass


class Server:
    def __init__(self, 
                 hostname="::",
                 port=19132, 
                 ipv=6, 
                 logger=None, 
                 logginglevel="INFO", 
                 game="MCPE", 
                 name="PieRakNet", 
                 game_protocol_version=729, 
                 version_name="1.21.30", 
                 max_player_count=20, 
                 modt="Powered by PieRakNet", 
                 game_mode="survival", 
                 game_mode_number=1, 
                 portv6=19133
                 ):
        if logger is None:
            logger = logging.getLogger("PieRakNet")
            logger.setLevel(getattr(logging, logginglevel.upper()))
            formatter = logging.Formatter('%(asctime)s [%(name)s - %(levelname)s] - %(message)s', "%H:%M:%S")
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        self.logger = logger
        self.hostname = hostname
        self.port = port
        self.ipv = ipv
        self.game = game
        self.name = name
        self.game_protocol_version = game_protocol_version
        self.version_name = version_name
        self.max_player_count = max_player_count
        self.server_id = "13253860892328930866"
        self.modt = modt
        self.game_mode = game_mode
        self.game_mode_number = game_mode_number
        self.portv6 = portv6
        self.raknet_protocol_version = 11
        self.guid = random.randint(0, sys.maxsize - 1)
        self.connections = []
        
        if self.ipv == 6:
            self.socket = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
            self.socket.bind((self.hostname, self.port, 0, 0))  # IPv6 bind
        else:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.socket.bind((self.hostname, self.port))  # IPv4 bind

        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 4096)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.start_time = time.time()
        self.magic = b'\x00\xff\xff\x00\xfe\xfe\xfe\xfe\xfd\xfd\xfd\xfd\x12\x34\x56\x78'
        self.running = False
        self.timeout = 15
        self.maxsize = 4096
        self.response_data = self.update_response_data()
        self.logger.info('Server initialized.')

    def send(self, data, address: tuple):
        if not isinstance(data, bytes):
            self.logger.debug(f"Encoding data to bytes: {data}")
            data = str(data).encode()
        self.logger.debug(f"Sending data to {address}")
        try:
            self.socket.sendto(data, address)
        except OSError as e:
            self.logger.error(f"Failed to send data to {address}: {e}")

    def update_response_data(self):
        player_count = len(self.connections)
        response_data = f"{self.game};{self.name};{self.game_protocol_version};{self.version_name};" \
               f"{player_count};{self.max_player_count};{self.server_id};{self.modt};" \
               f"{self.game_mode};{self.game_mode_number};{self.portv6};{self.port}"
        return response_data

    def get_connection(self, address):
        for connection in self.connections:
            if connection.address == address:
                return connection
        raise ConnectionNotFound()

    def add_connection(self, connection):
        if connection not in self.connections:
            self.connections.append(connection)
            self.update_response_data()
            self.logger.debug(f"Added connection: {connection} for address {connection.address}")
        else:
            self.logger.warning(f"Connection already exists: {connection.address}")

    def remove_connection(self, connection):
        if connection in self.connections:
            self.connections.remove(connection)
            self.update_response_data()
            self.logger.debug(f"Removed connection: {connection} for address {connection.address}")
        else:
            self.logger.warning(f"Connection not found: {connection.address}")

    def start(self):
        self.running = True
        self.logger.info(f"Server started on {self.hostname}:{self.port}! (IPv{'6' if self.ipv == 6 else '4'})")

        while self.running:
            ready_sockets, _, _ = select.select([self.socket], [], [], 0.05)
            if self.socket in ready_sockets:
                try:
                    data, client = self.socket.recvfrom(self.maxsize)
                except OSError as e:
                    self.logger.warning(f"OS error while receiving data: {e}")
                    continue
                if data:
                    self.handle_data(data, client)

    def handle_data(self, data, client):
        if data[0] in {ProtocolInfo.OFFLINE_PING, ProtocolInfo.OFFLINE_PING_OPEN_CONNECTIONS}:
            packet = OfflinePing(data)
            OfflinePingHandler.handle(packet, self, client)
        elif data[0] == ProtocolInfo.OPEN_CONNECTION_REQUEST_1:
            packet = OpenConnectionRequest1(data)
            OpenConnectionRequest1Handler.handle(packet, self, client)
        elif data[0] == ProtocolInfo.OPEN_CONNECTION_REQUEST_2:
            packet = OpenConnectionRequest2(data)
            OpenConnectionRequest2Handler.handle(packet, self, client)
        elif ProtocolInfo.FRAME_SET_0 <= data[0] <= ProtocolInfo.FRAME_SET_F:
            try:
                connection = self.get_connection(client)
                connection.handle(data)
            except ConnectionNotFound:
                self.logger.error(f"Connection not found for address {client}")

    def stop(self):
        self.running = False
        self.socket.close()
        self.logger.info('Server stopped.')



from pieraknet.packets.acknowledgement import Ack, Nack

class AckHandler:
    @staticmethod
    def handle(data: bytes, server, connection):
        """Handle received ACK packets."""
        ack_packet = Ack(data)
        ack_packet.decode()

        connection.logger.debug(f"Received ACK for sequence numbers: {ack_packet.sequence_numbers}")

        # Process the acknowledged sequence numbers
        for seq_num in ack_packet.sequence_numbers:
            if seq_num in connection.recovery_queue:
                del connection.recovery_queue[seq_num]
                connection.logger.debug(f"Acknowledged sequence number: {seq_num}")

    @staticmethod
    def send_ack(server, connection, sequence_number):
        """Send an ACK packet for the given sequence number."""
        ack_packet = Ack()
        ack_packet.sequence_numbers = [sequence_number]  # Sending a single ACK for the received sequence number

        ack_packet.encode()  # Ensure this is defined in the Ack class
        connection.send_data(ack_packet.getvalue())  # Send the ACK
        server.logger.debug(f"ACK sent for sequence number: {sequence_number}")

class NackHandler:
    @staticmethod
    def handle(data: bytes, server, connection):
        """ Handle received NACK packets. """
        nack_packet = Nack(data)
        nack_packet.decode()

        connection.logger.debug(f"Received NACK for sequence numbers: {nack_packet.sequence_numbers}")
        
        # Handle the negative acknowledgment (e.g., request retransmission)
        for seq_num in nack_packet.sequence_numbers:
            connection.logger.debug(f"Requesting retransmission for sequence number: {seq_num}")
            connection.nack_queue.append(seq_num)  # Add to NACK queue for retransmission



from pieraknet.server import Server

if __name__ == '__main__':
    server = Server(logginglevel = "INFO")
    try:
        server.start()
    except KeyboardInterrupt:
        server.logger.info('Stopping...')
        server.stop()


class FragmentedFrameHandler:
    @staticmethod
    def handle(frame, server, connection):
        server.logger.debug("New Fragmented Frame:")
        server.logger.debug(f"- Compound ID: {frame['compound_id']}")
        server.logger.debug(f"- Index: {frame['index']}")
        server.logger.debug(f"- Compound Size: {frame['compound_size']}")

        fragments = connection.fragmented_packets.setdefault(frame['compound_id'], {})
        fragments[frame['index']] = frame

        if len(fragments) == frame['compound_size']:
            server.logger.debug("Reassembling fragmented frame:")
            body = b''.join(fragments[i]['body'] for i in range(frame['compound_size']))
            new_frame = {
                'flags': 0,
                'length_in_bits': len(body) * 8,
                'body': body,
                'reliable_frame_index': 0,
                'sequenced_frame_index': 0,
                'ordered_frame_index': 0,
                'order_channel': 0,
                'compound_size': 0,
                'compound_id': 0,
                'index': 0
            }
            del connection.fragmented_packets[frame['compound_id']]
            connection.handle_frame(new_frame)



from pieraknet.packets.frame_set import FrameSetPacket
from pieraknet.handlers.frame import FrameHandler
from pieraknet.handlers.fragmented_frame import FragmentedFrameHandler
from pieraknet.handlers.acknowledgement import AckHandler

class FrameSetHandler:
    @staticmethod
    def handle(data: bytes, server, connection):
        frame_set_packet = FrameSetPacket(server)
        frame_set_packet.decode(data)
        incoming_sequence_number = frame_set_packet.sequence_number

        server.logger.debug("New Packet:")
        server.logger.debug(f"- Packet ID: {frame_set_packet.packet_id}")
        server.logger.debug(f"- Packet Name: Frame Set")
        server.logger.debug(f"- Incoming Sequence Number: {incoming_sequence_number}")

        # Process if the sequence number has not been processed
        if incoming_sequence_number not in connection.client_sequence_numbers:
            connection.client_sequence_numbers.add(incoming_sequence_number)

            # Send ACK for the received sequence number
            AckHandler.send_ack(server=server, connection=connection, sequence_number=incoming_sequence_number)
            connection.ack_queue.append(incoming_sequence_number)

            # Update the client sequence number
            connection.client_sequence_number = incoming_sequence_number

            # Process each frame in the packet
            for frame in frame_set_packet.frames:
                FrameSetHandler.process_frame(frame, server, connection)

    @staticmethod
    def process_frame(frame, server, connection):
        if frame['flags'] & 0x08:  # Check if it's a fragmented frame
            server.logger.debug("Handling fragmented frame...")
            FragmentedFrameHandler.handle(frame, server, connection)
        else:
            server.logger.debug("Handling frame...")
            FrameHandler.handle(frame, server, connection)



from pieraknet.protocol_info import ProtocolInfo
from pieraknet.handlers.online_ping import OnlinePingHandler
from pieraknet.handlers.game_packet import GamePacketHandler
from pieraknet.handlers.disconnect import DisconnectHandler
from pieraknet.handlers.unknown_packet import UnknownPacketHandler

class EstablishedConnectionHandler:
    @staticmethod
    def handle(frame, server, connection):
        packet_type = frame['body'][0]

        server.logger.debug("Established Connection Packet:")
        server.logger.debug(f"- Packet Type: {packet_type}")
        
        if packet_type == ProtocolInfo.ONLINE_PING:
            OnlinePingHandler.process_online_ping(frame, server, connection)
        elif packet_type == ProtocolInfo.GAME_PACKET:
            GamePacketHandler.handle(frame, server, connection)
        elif packet_type == ProtocolInfo.DISCONNECT:
            DisconnectHandler.handle(frame['body'], server, connection)
        else:
            UnknownPacketHandler.handle(frame['body'], server, connection)



class GamePacketHandler:
    @staticmethod
    def handle(frame, server, connection):
        server.logger.debug("Game Packet:")
        server.logger.debug(f"- Packet Body Length: {len(frame['body'])}")

        if hasattr(server, "interface") and hasattr(server.interface, "on_game_packet"):
            server.interface.on_game_packet(frame, connection)



from pieraknet.buffer import Buffer
from pieraknet.packets.packet import Packet
from pieraknet.protocol_info import ProtocolInfo

class FrameSetPacket(Packet):
    def __init__(self, server=None, data: bytes = b''):
        super().__init__(data)
        self.server = server
        self.packet_id = ProtocolInfo.FRAME_SET
        self.sequence_number = 0
        self.frames = []

    def decode(self, data):
        buffer = Buffer(data)
        self.packet_id = buffer.read_byte()
        self.sequence_number = buffer.read_uint24le()

        while not buffer.feos():
            frame = self.decode_frame(buffer)
            self.frames.append(frame)

    def decode_frame(self, buffer):

        frame = {
            'flags': buffer.read_byte(),
            'length_in_bits': buffer.read_unsigned_short(),
            'reliable_frame_index': 0,
            'sequenced_frame_index': 0,
            'ordered_frame_index': 0,
            'order_channel': 0,
            'compound_size': 0,
            'compound_id': 0,
            'index': 0,
            'body': b''
        }

        reliability_type = (frame['flags'] >> 5) & 0x07
        is_fragmented = (frame['flags'] >> 4) & 0x01

        if reliability_type in {2, 3, 4, 6, 7}:
            frame['reliable_frame_index'] = buffer.read_uint24le()

        if reliability_type in {1, 4}:
            frame['sequenced_frame_index'] = buffer.read_uint24le()

        if reliability_type in {1, 3, 4, 7}:
            frame['ordered_frame_index'] = buffer.read_uint24le()
            frame['order_channel'] = buffer.read_byte()

        if is_fragmented:
            frame['compound_size'] = buffer.read_int()
            frame['compound_id'] = buffer.read_short()
            frame['index'] = buffer.read_int()

        body_length = (frame['length_in_bits'] + 7) // 8
        frame['body'] = buffer.read(body_length)

        return frame

    def encode(self):
        buffer = Buffer()
        buffer.write_byte(self.packet_id)
        buffer.write_uint24le(self.sequence_number)

        for frame in self.frames:
            self.encode_frame(buffer, frame)

        return buffer.getvalue()

    def encode_frame(self, buffer, frame):
        buffer.write_byte(frame['flags'])
        buffer.write_unsigned_short(frame['length_in_bits'])

        reliability_type = (frame['flags'] >> 5) & 0x07
        is_fragmented = (frame['flags'] >> 4) & 0x01

        if reliability_type in {2, 3, 4, 6, 7}:
            buffer.write_uint24le(frame['reliable_frame_index'])

        if reliability_type in {1, 4}:
            buffer.write_uint24le(frame['sequenced_frame_index'])

        if reliability_type in {1, 3, 4, 7}:
            buffer.write_uint24le(frame['ordered_frame_index'])
            buffer.write_byte(frame['order_channel'])

        if is_fragmented:
            buffer.write_int(frame['compound_size'])
            buffer.write_short(frame['compound_id'])
            buffer.write_int(frame['index'])

        buffer.write(frame['body'])

    def create_frame(self, body: bytes, flags: int = 0):
        frame = {
            'flags': flags,
            'length_in_bits': len(body) * 8,
            'body': body,
            'reliable_frame_index': 0,
            'sequenced_frame_index': 0,
            'ordered_frame_index': 0,
            'order_channel': 0,
            'compound_size': 0,
            'compound_id': 0,
            'index': 0
        }
        self.frames.append(frame)

    def set_sequence_number(self, sequence_number: int):
        self.sequence_number = sequence_number


from pieraknet.packets.online_ping import OnlinePing
from pieraknet.packets.online_pong import OnlinePong
from pieraknet.packets.frame_set import FrameSetPacket
import time

class OnlinePingHandler:
    @staticmethod
    def handle(packet: OnlinePing, server):
        packet.decode()

        server.logger.debug("New Packet:")
        server.logger.debug(f"- Packet ID: {packet.PACKET_ID}")
        server.logger.debug(f"- Packet Name: Online Ping")
        server.logger.debug(f"- Client Timestamp: {packet.client_timestamp}")

        new_packet = OnlinePong()
        new_packet.client_timestamp = packet.client_timestamp
        new_packet.server_timestamp = int(time.time() * 1000)

        new_packet.encode()
        return new_packet.getvalue()

    def create_online_ping(server, connection):
        new_ping = OnlinePing()
        new_ping.client_timestamp = int(time.time() * 1000)
        new_ping.encode()
        OnlinePingPacket = new_ping.getvalue()

        frame_set_packet = FrameSetPacket(server)
        frame_set_packet.sequence_number = connection.server_sequence_number
        frame_set_packet.create_frame(OnlinePingPacket, flags=0x00)

        connection.send_data(frame_set_packet.encode()) 
    
    def process_online_ping(frame, server, connection):

        OnlinePingPacket = OnlinePing(frame['body'])

        OnlinePongPacket = OnlinePingHandler.handle(OnlinePingPacket, server)

        frame_set_packet = FrameSetPacket(server)
        frame_set_packet.sequence_number = connection.server_sequence_number
        frame_set_packet.create_frame(OnlinePongPacket, flags=0x00)

        connection.send_data(frame_set_packet.encode())



from pieraknet.protocol_info import ProtocolInfo
from pieraknet.handlers.disconnect import DisconnectHandler
from pieraknet.handlers.established_connection import EstablishedConnectionHandler

class FrameHandler:
    @staticmethod
    def handle(frame, server, connection):
        server.logger.debug("New Frame:")
        server.logger.debug(f"- Frame Flags: {frame['flags']}")
        server.logger.debug(f"- Frame Body Length: {len(frame['body'])}")

        if not (isinstance(frame, dict) and 'flags' in frame and 'body' in frame):
            server.logger.error("Invalid frame structure")
            return

        if not connection.connected:
            connection.handle_connection_requests(frame)
        else:
            EstablishedConnectionHandler.handle(frame, server, connection)

        if frame['body'] and frame['body'][0] == ProtocolInfo.DISCONNECT:
            DisconnectHandler.handle(frame['body'], server, connection)



import time
import collections
from pieraknet.packets.frame_set import FrameSetPacket
from pieraknet.protocol_info import ProtocolInfo
from pieraknet.handlers.connection_request import ConnectionRequestHandler
from pieraknet.handlers.new_incoming_connection import NewIncomingConnectionHandler
from pieraknet.handlers.acknowledgement import AckHandler, NackHandler
from pieraknet.handlers.frame_set import FrameSetHandler
from pieraknet.handlers.disconnect import DisconnectHandler
from pieraknet.handlers.packet_loss import PacketLossHandler
from pieraknet.handlers.online_ping import OnlinePingHandler

class Connection:
    def __init__(self, server, address):
        self.server = server
        self.address = address
        self.connected = False
        self.client_sequence_number = 0
        self.server_sequence_number = 0
        self.ack_queue = collections.deque()  # Optimized for append/pop operations
        self.nack_queue = collections.deque()
        self.recovery_queue = collections.OrderedDict()  # Keep recovery packets in order
        self.client_sequence_numbers = set()  # Efficient lookups for incoming sequence numbers
        self.fragmented_packets = {}
        self.logger = server.logger
        self.logger.debug(f"Created connection: {self} for address {self.address}")

    def disconnect(self):
        self.server.logger.debug(f"Disconnecting {self.address}")
        self.connected = False
        self.ack_queue.clear()
        self.nack_queue.clear()
        self.recovery_queue.clear()
        self.server.remove_connection(self)

    def handle(self, data):
        """ Handle incoming packets based on packet_id """
        packet_id = data[0]

        handler_map = {
            ProtocolInfo.ACK: AckHandler.handle,
            ProtocolInfo.NACK: NackHandler.handle,
            ProtocolInfo.DISCONNECT: DisconnectHandler.handle,
            **{pid: FrameSetHandler.handle for pid in range(ProtocolInfo.FRAME_SET_0, ProtocolInfo.FRAME_SET_F + 1)},
        }

        if handler := handler_map.get(packet_id):
            handler(data, self.server, self)
        else:
            self.server.logger.error(f"Unknown packet ID: {packet_id}")

    def process_retransmissions(self):
        """ Retry sending unacknowledged packets """
        current_time = time.time()
        to_resend = []

        for seq_num, (packet, timestamp) in list(self.recovery_queue.items()):
            if current_time - timestamp > self.server.timeout:
                to_resend.append((seq_num, packet))
                self.logger.debug(f"Resending packet with sequence number {seq_num}")

        for seq_num, packet in to_resend:
            self.send_data(packet)

    def update(self):
        """ Periodically check for lost packets and handle ACK/NACK """
        self.acknowledge()
        self.process_retransmissions()

    def handle_packet_loss(self, incoming_sequence_number):
        """ Detect and handle lost packets """
        missing_packets = range(self.client_sequence_number + 1, incoming_sequence_number)
        if missing_packets:
            self.nack_queue.extend(missing_packets)
            PacketLossHandler.handle(incoming_sequence_number, self.server, self)

    def handle_connection_requests(self, frame):
        packet_type = frame['body'][0]
        if packet_type == ProtocolInfo.CONNECTION_REQUEST:
            OnlinePingHandler.create_online_ping(self.server, self)
            connection_packet = ConnectionRequestHandler.handle(frame['body'], self.server, self)
            # Crear un FrameSetPacket
            frame_set_packet = FrameSetPacket(self.server)
            frame_set_packet.sequence_number = self.client_sequence_number
            frame_set_packet.create_frame(connection_packet, flags=0x64)

            # Codificar y enviar directamente sin usar Buffer
            self.send_data(frame_set_packet.encode())
            self.connected = True
        elif packet_type == ProtocolInfo.NEW_INCOMING_CONNECTION:
            NewIncomingConnectionHandler.handle(frame['body'], self.server, self)

    def send_data(self, data):
        self.server.send(data, self.address)
        # if its a frame set packet
        if ProtocolInfo.FRAME_SET_0 <= data[0] <= ProtocolInfo.FRAME_SET_F:
            self.recovery_queue[self.server_sequence_number] = (data, time.time())
            self.server_sequence_number += 1

    def acknowledge(self):
        """ Send accumulated ACKs """
        if self.ack_queue:
            ack_packet = AckHandler.create_ack_packet(list(self.ack_queue))
            self.send_data(ack_packet)
            self.ack_queue.clear()

        if self.nack_queue:
            nack_packet = NackHandler.create_nack_packet(list(self.nack_queue))
            self.send_data(nack_packet)
            self.nack_queue.clear()
